/* $Id$

This file is part of libmspcore
Copyright © 2006-2007 Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/
#ifndef MSP_CORE_GETOPT_H_
#define MSP_CORE_GETOPT_H_

#include <sstream>
#include <string>
#include <vector>
#include "except.h"

namespace Msp {

class GetOpt
{
public:
	enum ArgType
	{
		NO_ARG,
		OPTIONAL_ARG,
		REQUIRED_ARG
	};
	
	class OptBase
	{
	public:
		OptBase           &set_help(const std::string &h) { help=h; return *this; }
		char              get_short() const               { return shrt; }
		const std::string &get_long() const               { return lng; }
		ArgType           get_arg_type() const            { return arg_type; }
		const std::string &get_help() const               { return help; }
		unsigned          get_seen_count() const          { return seen_count; }
		virtual void      process()=0;
		virtual void      process(const std::string &)=0;
		virtual ~OptBase() { }
	protected:
		char        shrt;
		std::string lng;
		ArgType     arg_type;
		unsigned    seen_count;
		std::string help;

		OptBase(char s, const std::string &l, ArgType a): shrt(s), lng(l), arg_type(a), seen_count(0) { }
	};

	const std::vector<std::string> &get_args() const { return args; }

	template<typename T>
	OptBase &add_option(char s, const std::string &l, T &d, ArgType a=NO_ARG)
	{ opts.push_back(new Option<T>(s, l, d, a)); return *opts.back(); }
	
	template<typename T>
	OptBase &add_option(char s, const std::string &l, std::list<T> &d, ArgType a=REQUIRED_ARG)
	{ opts.push_back(new ListOption<std::list<T> >(s, l, d, a)); return *opts.back(); }
	
	template<typename T>
	OptBase &add_option(const std::string &l, T &d, ArgType a)
	{ return add_option(0, l, d, a); }

	std::string generate_usage(const std::string &) const;
	std::string generate_help() const;
	void operator()(unsigned, const char *const *);

	~GetOpt();
private:
	template<typename T>
	class Option: public OptBase
	{
	public:
		Option(char s, const std::string &l, T &d, ArgType a): OptBase(s, l, a), data(d) { }

		virtual void process()
		{
			if(arg_type==REQUIRED_ARG)
				throw UsageError("--"+lng+" requires an argument");
			process_();
			++seen_count;
		}

		virtual void process(const std::string &a)
		{
			if(arg_type==NO_ARG)
				throw UsageError("--"+lng+" takes no argument");

			T tmp;
			std::istringstream ss(a);
			ss>>tmp;
			if(ss.fail())
				throw UsageError("Invalid argument for --"+lng);

			data=tmp;
			++seen_count;
		}
	private:
		T &data;

		void process_() { }
	};

	template<typename T>
	class ListOption: public OptBase
	{
	public:
		ListOption(char s, const std::string &l, T &d, ArgType a): OptBase(s, l, a), data(d)
		{ if(arg_type!=REQUIRED_ARG) throw Exception("ListOption with arg_type!=REQUIRED makes no sense"); }

		virtual void process()
		{
			throw UsageError("--"+lng+" requires an argument");
		}

		virtual void process(const std::string &a)
		{
			typename T::value_type tmp;
			std::istringstream ss(a);
			ss>>tmp;
			if(ss.fail())
				throw UsageError("Invalid argument for --"+lng);

			data.push_back(tmp);
			++seen_count;
		}
	private:
		T &data;
	};

	std::list<OptBase *>   opts;
	std::vector<std::string> args;

	OptBase &get_option(char);
	OptBase &get_option(const std::string &);
	unsigned process_long(const char *const *);
	unsigned process_short(const char *const *);
};

template<> inline void GetOpt::Option<bool>::process_()     { data=true; }
template<> inline void GetOpt::Option<unsigned>::process_() { ++data; }

} // namespace Msp

#endif
